from typing import TYPE_CHECKING, List

from boto3 import resource
from opentelemetry import trace

from app.core.config import settings

if TYPE_CHECKING:
    from mypy_boto3_s3.service_resource import BucketPolicy, ObjectSummary, S3ServiceResource
else:
    S3ServiceResource = object
    BucketPolicy = object
    ObjectSummary = object


tracer = trace.get_tracer_provider().get_tracer(__name__)

s3_resource: S3ServiceResource = resource(
    service_name="s3",
    endpoint_url=str(settings.OBJECT_GATEWAY_URI)[:-1],
    aws_access_key_id=settings.BUCKET_CEPH_ACCESS_KEY,
    aws_secret_access_key=settings.BUCKET_CEPH_SECRET_KEY,
    verify=str(settings.OBJECT_GATEWAY_URI).startswith("https"),
)


def get_s3_bucket_policy(s3: S3ServiceResource, bucket_name: str) -> BucketPolicy:
    with tracer.start_as_current_span("s3_get_bucket_policy") as span:
        span.set_attribute("bucket_name", bucket_name)
        s3_policy = s3.Bucket(bucket_name).Policy()
        s3_policy.load()
        return s3_policy


def put_s3_bucket_policy(s3: S3ServiceResource, bucket_name: str, policy: str) -> None:
    with tracer.start_as_current_span("s3_put_bucket_policy") as span:
        span.set_attribute("bucket_name", bucket_name)
        s3.Bucket(bucket_name).Policy().put(Policy=policy)


def get_s3_bucket_objects(s3: S3ServiceResource, bucket_name: str) -> List[ObjectSummary]:
    with tracer.start_as_current_span("s3_get_object_meta_data") as span:
        span.set_attribute("bucket_name", bucket_name)
        return list(s3.Bucket(bucket_name).objects.all())
