from enum import Enum, unique
from typing import Optional, Sequence

from clowmdb.models import Bucket
from clowmdb.models import BucketPermission as BucketPermissionDB
from opentelemetry import trace
from sqlalchemy import delete, func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession

from app.crud import DuplicateError
from app.otlp import start_as_current_span_async
from app.schemas.bucket import BucketIn as BucketInSchema

tracer = trace.get_tracer_provider().get_tracer(__name__)


class CRUDBucket:
    @unique
    class BucketType(str, Enum):
        """
        Enumeration for the type of buckets to fetch from the DB

        OWN: Only fetch buckets that the user owns
        PERMISSION: Only fetch foreign buckets that the user has access to
        ALL: Fetch all buckets that the user has access to
        """

        OWN: str = "OWN"
        ALL: str = "ALL"
        PERMISSION: str = "PERMISSION"

    @staticmethod
    @start_as_current_span_async("db_get_bucket", tracer=tracer)
    async def get(db: AsyncSession, bucket_name: str) -> Optional[Bucket]:
        """
        Get a bucket by its name.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        bucket_name : str
            Name of the Bucket to get from database.

        Returns
        -------
        bucket : Bucket | None
            Returns the bucket if it exists, None otherwise.
        """
        stmt = select(Bucket).where(Bucket.name == bucket_name)
        trace.get_current_span().set_attributes({"bucket_name": bucket_name, "sql_query": str(stmt)})
        return await db.scalar(stmt)

    @staticmethod
    @start_as_current_span_async("db_list_all_buckets", tracer=tracer)
    async def get_all(db: AsyncSession) -> Sequence[Bucket]:
        stmt = select(Bucket)
        trace.get_current_span().set_attribute("sql_query", str(stmt))
        return (await db.scalars(stmt)).all()

    @staticmethod
    @start_as_current_span_async("db_list_buckets_for_user", tracer=tracer)
    async def get_for_user(db: AsyncSession, uid: str, bucket_type: BucketType = BucketType.ALL) -> Sequence[Bucket]:
        """
        Get all buckets for a user. Depending on the `bucket_type`, the user is either owner of the bucket or has
        permission for the bucket

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        uid : str
            UID of a user.
        bucket_type : BucketType, default BucketType.ALL


        Returns
        -------
        buckets : List[clowmdb.models.Bucket]
            A list of all buckets where the given user has READ permissions for.

        Notes
        -----
        SQL Query own buckets:
        ```
        SELECT bucket.name, bucket.description, bucket.public, bucket.owner_id
        FROM bucket
        WHERE bucket.owner_id = %s
        ```

        SQL Query all buckets that the user has access to:
        ```
        SELECT bucket.name, bucket.description, bucket.public, bucket.owner_id
        FROM bucket
        WHERE bucket.owner_id = %s OR (EXISTS
            (SELECT 1 FROM bucketpermission
            WHERE bucket.name = bucketpermission.bucket_name AND bucketpermission.user_id = %s
                AND(UNIX_TIMESTAMP() >= bucketpermission.`from` 0 OR bucketpermission.`from` IS NULL)
                AND(UNIX_TIMESTAMP() <= bucketpermission.`to` >= 0 OR bucketpermission.`to` IS NULL)))
        ```

        SQL Query only foreign buckets where user has permission to
        ```
        SELECT bucket.name, bucket.description, bucket.public, bucket.owner_id
        FROM bucket
        WHERE (EXISTS
            (SELECT 1 FROM bucketpermission
            WHERE bucket.name = bucketpermission.bucket_name AND bucketpermission.user_id = %s
                AND(UNIX_TIMESTAMP() >=  bucketpermission.`from` <= 0 OR bucketpermission.`from` IS NULL)
                AND(UNIX_TIMESTAMP() <=  bucketpermission.`to` >= 0 OR bucketpermission.`to` IS NULL)))
        ```
        """
        stmt = select(Bucket)
        if bucket_type == CRUDBucket.BucketType.ALL:
            stmt = stmt.where(
                or_(
                    Bucket.owner_id == uid,
                    Bucket.permissions.any(BucketPermissionDB.user_id == uid)
                    .where(
                        or_(
                            func.UNIX_TIMESTAMP() >= BucketPermissionDB.from_,
                            BucketPermissionDB.from_ == None,  # noqa:E711
                        )
                    )
                    .where(
                        or_(
                            func.UNIX_TIMESTAMP() <= BucketPermissionDB.to,
                            BucketPermissionDB.to == None,  # noqa:E711
                        )
                    ),
                )
            )
        elif bucket_type == CRUDBucket.BucketType.OWN:
            stmt = stmt.where(Bucket.owner_id == uid)
        else:
            stmt = stmt.where(
                Bucket.permissions.any(BucketPermissionDB.user_id == uid)
                .where(
                    or_(
                        func.UNIX_TIMESTAMP() >= BucketPermissionDB.from_,
                        BucketPermissionDB.from_ == None,  # noqa:E711
                    )
                )
                .where(
                    or_(
                        func.UNIX_TIMESTAMP() <= BucketPermissionDB.to,
                        BucketPermissionDB.to == None,  # noqa:E711
                    )
                ),
            )
        trace.get_current_span().set_attributes({"sql_query": str(stmt), "uid": uid, "bucket_type": bucket_type.name})
        return (await db.scalars(stmt)).all()

    @staticmethod
    @start_as_current_span_async("db_create_bucket", tracer=tracer)
    async def create(db: AsyncSession, bucket_in: BucketInSchema, uid: str) -> Bucket:
        """
        Create a bucket for a given user.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket_in : app.schemas.bucket.BucketIn
            All relevant information for a new bucket.
        uid : str
            UID of the owner for the new bucket.

        Returns
        -------
        bucket : clowmdb.models.Bucket
            Returns the created bucket.
        """
        bucket = Bucket(**bucket_in.model_dump(), owner_id=uid)
        current_span = trace.get_current_span()
        current_span.set_attribute("bucket_name", bucket.name)
        if await CRUDBucket.get(db, bucket.name) is not None:
            raise DuplicateError(f"Bucket {bucket.name} exists already")
        db.add(bucket)
        await db.commit()
        await db.refresh(bucket)
        return bucket

    @staticmethod
    @start_as_current_span_async("db_delete_bucket", tracer=tracer)
    async def delete(db: AsyncSession, bucket_name: str) -> None:
        """
        Delete a specific bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket_name : string
            The name of the bucket to delete.
        """
        stmt = delete(Bucket).where(Bucket.name == bucket_name)
        trace.get_current_span().set_attributes({"sql_query": str(stmt), "bucket_name": bucket_name})
        await db.execute(stmt)
        await db.commit()
