import re
from typing import Optional

from clowmdb.models import Bucket
from pydantic import BaseModel, ConfigDict, Field, field_validator

ip_like_regex = re.compile(r"^(\d+\.){3}\d+$")


class _BaseBucket(BaseModel):
    """
    Base Schema for a bucket.
    """

    name: str = Field(
        ...,
        examples=["test-bucket"],
        description="Name of the bucket",
        min_length=3,
        max_length=63,
        pattern=r"^([a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)*[a-z0-9](?:[a-z0-9-]*[a-z0-9])?$",
        # https://docs.ceph.com/en/latest/radosgw/s3/bucketops/#constraints
    )
    description: str = Field(
        ...,
        examples=["This is a sample description of a bucket"],
        description="Description of the bucket",
        min_length=32,
        max_length=2**16,
    )

    @field_validator("name")
    @classmethod
    def name_is_not_an_ip_address(cls, name: str) -> str:
        if ip_like_regex.search(name):
            raise ValueError("no IP address as bucket name")
        return name


class BucketIn(_BaseBucket):
    """
    Schema for creating a new bucket.
    """


class BucketOut(_BaseBucket):
    """
    Schema for answering a request with a bucket.
    """

    created_at: int = Field(
        ...,
        examples=[1640991600],  # 01.01.2022 00:00
        description="Time when the bucket was created as UNIX timestamp",
    )
    owner: str = Field(..., description="UID of the owner", examples=["28c5353b8bb34984a8bd4169ba94c606"])
    num_objects: int = Field(..., description="Number of Objects in this bucket", examples=[6])
    size: int = Field(..., description="Total size of objects in this bucket in bytes", examples=[3256216])
    owner_constraint: Optional[Bucket.Constraint] = Field(None, description="Constraint for the owner of the bucket")
    description: str = Field(
        ...,
        description="Description of the bucket",
    )
    model_config = ConfigDict(from_attributes=True)
